package nstaclient

import (
	"context"
	"fmt"
	"log"
	"os"
	"os/signal"
	"path/filepath"
	"syscall"
	"time"

	"github.com/it-novum/nsta/checkmk"
	"github.com/it-novum/nsta/gearman"
	"github.com/it-novum/nsta/httpclient"
	"github.com/it-novum/nsta/nstautil"
)

type ClientPush struct {
	ClientSettings

	gearman          *gearman.Client
	client           *httpclient.NSTAHTTPClient
	nagiosConfigPath string
	cmk              checkmk.Checkmk
	cmdBackend       *checkmk.LocalCommandBackend

	lastExport time.Time

	toMaster gearman.SyncMessageFIFO
}

func (c *ClientPush) StartForeground() error {
	ctx, cancel := context.WithCancel(context.Background())

	c.nagiosConfigPath = filepath.Join(c.NagiosPath, "etc")
	c.lastExport = time.Time{}
	c.cmdBackend = &checkmk.LocalCommandBackend{
		Timeout:       time.Second * 120,
		SatelliteName: "Local",
	}

	if httpClient, err := httpclient.New(ctx, c.Endpoint, c.SatelliteID, c.APIKey, c.Proxy, c.VerifyCertificate); err != nil {
		cancel()
		return err
	} else {
		c.client = httpClient
	}

	c.cmk.SatelliteName = "Local"
	c.toMaster.Start()

	c.gearman = &gearman.Client{
		ClientOptions: gearman.ClientOptions{
			ClientType:     gearman.Satellite,
			ClientName:     "Local",
			Address:        c.GearmanAddress,
			MessageReceive: c.toMaster.Direct(),
		},
	}
	c.gearman.Start(ctx)

	defer cancel()
	defer c.toMaster.Close()
	defer c.gearman.Stop()
	c.run(ctx)

	return nil
}

func (c *ClientPush) processMessage(ctx context.Context) error {
	msgs := c.toMaster.GetAll(8192)
	response, err := c.client.SendMessages(msgs, c.lastExport)
	if err != nil {
		// Error while sending data - put messages back to FIFO cache
		c.toMaster.PushAll(msgs, true)
		log.Println(&Error{Name: "Local", Err: err})
		return err
	}
	if response != nil {
		log.Println(&Error{Name: "Local", Message: fmt.Sprint("sent ", len(msgs), " and received ", len(response.Messages), " messages")})

		for _, r := range response.Messages {
			switch r.Type {
			case gearman.MessageCheckmk:

				// READ!! https://stackoverflow.com/a/30183893
				go func(gearmanMessageCopy *gearman.Message) {
					response, err := c.cmk.ProcessMessage(c.cmdBackend, gearmanMessageCopy).Serialize()
					if err != nil {
						log.Println(&Error{Name: "Local", Err: err})
						return
					}
					c.toMaster.Push(&gearman.Message{
						Type:   gearman.MessageCheckmk,
						String: response,
					})
				}(r)
			case gearman.MessageTimesync:
				go func() {
					ts := gearman.TimesyncMessage{}
					if err := json.Unmarshal([]byte(r.String), &ts); err != nil {
						log.Println(&Error{Name: "Local", Message: "received invalid timesync message", Err: err})
					} else {
						if _, err := c.cmdBackend.RunLocalCommand(ctx, "", []string{"sudo", "date", "-s", fmt.Sprint("@", ts.SetTime.Unix())}, map[string]string{}); err != nil {
							log.Println(err)
						} else {
							log.Println(&Error{Name: "Local", Message: "successful updated time"})
						}
					}
				}()
			default:
				c.gearman.SendMessage(r)
			}
		}
		if !response.LastExport.IsZero() {
			if response.LastExport.After(c.lastExport) {
				//Master did an export. Send http request to master server to query the new config as base64 tar.gz string
				export, err := c.client.ExportExchange(c.lastExport, "")
				if err == nil && export != nil {
					if export.TARBase64 == "" {
						log.Println(&Error{Name: "Local", Message: "requst new export"})
					} else {
						if err = nstautil.InstallExport(c.nagiosConfigPath, export.TARBase64); err != nil {
							log.Println(&Error{Name: "Local", Err: err})
						} else {
							c.lastExport = export.LastExport
						}

					}
				}
			}
		}
	}

	return nil
}

func (c *ClientPush) run(ctx context.Context) {
	var notify chan struct{}
	emptyChan := make(chan struct{})

	// Bind linux kernel signal handler
	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, syscall.SIGINT, syscall.SIGTERM)

	ticker := time.NewTicker(c.PushInterval)
	defer ticker.Stop()
	lastError := false

	for {
		if !c.InstantPush || lastError {
			notify = emptyChan
		} else {
			notify = c.toMaster.Notify()
		}
		select {
		case <-ctx.Done():
			return
		case <-sigs:
			// Exit on SIGINT / SIGTERM
			return
		case <-notify:
			// If InstantPush is enabled
			// Gets signal from FIFO
			err := c.processMessage(ctx)
			if err != nil {
				lastError = true
			} else {
				lastError = false
			}
		case <-ticker.C:
			// Push results every x seconds to master
			err := c.processMessage(ctx)
			if err != nil {
				lastError = true
			} else {
				lastError = false
			}
		}
	}
}
